/* This file is Copyright 2002 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include "reflector/ReflectServer.h"
#include "reflector/DumbReflectSession.h"
#include "reflector/StorageReflectSession.h"
#include "reflector/FilterSessionFactory.h"
#include "reflector/RateLimitSessionIOPolicy.h"
#include "system/GlobalMemoryAllocator.h"
#include "system/SetupSystem.h"
#include "util/StringTokenizer.h"
#include "regex/StringMatcher.h"

using namespace muscle;

#define DEFAULT_MUSCLED_PORT 2960

// Aux method; main() without the global stuff.  This is a good method to
// call if you already have the global stuff set up the way you like it.
// The third argument can be passed in as NULL, or point to a UsageLimitProxyMemoryAllocator object
int muscledmainAux(int argc, char ** argv, void * cookie)
{
   UsageLimitProxyMemoryAllocator * usageLimitAllocator = (UsageLimitProxyMemoryAllocator*)cookie;

   uint32 maxBytes           = MUSCLE_NO_LIMIT;
   uint32 maxNodesPerSession = MUSCLE_NO_LIMIT;
   uint32 maxReceiveRate     = MUSCLE_NO_LIMIT;
   uint32 maxSendRate        = MUSCLE_NO_LIMIT;
   uint32 maxCombinedRate    = MUSCLE_NO_LIMIT;
   uint32 maxMessageSize     = MUSCLE_NO_LIMIT;
   uint32 maxSessions        = MUSCLE_NO_LIMIT;
   uint32 maxSessionsPerHost = MUSCLE_NO_LIMIT;

   Queue<uint16> ports;
   Queue<String> bans;
   Queue<String> requires;
   Message tempPrivs;

   // Parse new-style arguments (e.g. "./muscled port=2960 maxmem=100")
   for (int i=1; i<argc; i++)
   {
      StringTokenizer tok(argv[i], (const char *)"=");
      const char * keyword = tok.GetNextToken();

      if (keyword == NULL) keyword = argv[i];
      if (keyword)
      {
         const char * value = tok.GetNextToken();
         if (value == NULL) value = "";

         while(keyword[0] == '-') keyword++;  // stupid Unix geeks anyway
         if (strcmp(keyword, "help") == 0)
         {
            Log(MUSCLE_LOG_INFO, "Usage:  muscled [port=%u] [display=lvl] [log=lvl]\n", DEFAULT_MUSCLED_PORT); 
            Log(MUSCLE_LOG_INFO, "                [maxmem=megs] [maxnodespersession=num]\n");
            Log(MUSCLE_LOG_INFO, "                [ban=ippattern] [require=ippattern]\n");
            Log(MUSCLE_LOG_INFO, "                [privban=ippattern] [privunban=ippattern]\n");
            Log(MUSCLE_LOG_INFO, "                [privkick=ippattern] [privall=ippattern]\n");
            Log(MUSCLE_LOG_INFO, "                [maxsendrate=kBps] [maxreceiverate=kBps]\n");
            Log(MUSCLE_LOG_INFO, "                [maxcombinedrate=kBps] [maxmessagesize=k]\n");
            Log(MUSCLE_LOG_INFO, "                [maxsessions=num] [maxsessionsperhost=num]\n");
            Log(MUSCLE_LOG_INFO, "                [localhost=ipaddress] [daemon]\n");
            Log(MUSCLE_LOG_INFO, " - port may be any number between 1 and 65536\n");
            Log(MUSCLE_LOG_INFO, " - lvl is: none, critical, errors, warnings, info, debug, or trace.\n");
            Log(MUSCLE_LOG_INFO, " - maxmem is the max megabytes of memory the server may use (default=unlimited)\n");
            Log(MUSCLE_LOG_INFO, " - You may also put one or more ban=<pattern> arguments in.\n");
            Log(MUSCLE_LOG_INFO, "   Each pattern specifies one or more IP addresses to\n");
            Log(MUSCLE_LOG_INFO, "   disallow connections from, e.g. ban=192.168.*.*\n");
            Log(MUSCLE_LOG_INFO, " - You may put one or more require=<pattern> arguments in.\n");
            Log(MUSCLE_LOG_INFO, "   If any of these are present, then only IP addresses that match\n");
            Log(MUSCLE_LOG_INFO, "   at least one of them will be allowed to connect.\n");
            Log(MUSCLE_LOG_INFO, " - To assign privileges, specify one of the following:\n");
            Log(MUSCLE_LOG_INFO, "   privban=<pattern>, privunban=<pattern>,\n");
            Log(MUSCLE_LOG_INFO, "   privkick=<pattern> or privall=<pattern>.\n");
            Log(MUSCLE_LOG_INFO, "   privall assigns all privileges to the matching IP addresses.\n");
            Log(MUSCLE_LOG_INFO, " - If daemon is specified, muscled will run as a background process\n");
            return(5);
         }
         else 
         {
            if (strcmp(keyword, "port") == 0) 
            {
               int port = atoi(value);
               if (port > 0) ports.AddTail((uint16)port);
            }
            else if (strcmp(keyword, "daemon") == 0)
            {
               LogTime(MUSCLE_LOG_INFO, "Spawning off a muscled daemon-child...\n");
               if (BecomeDaemonProcess("/") != B_NO_ERROR)
               {
                  LogTime(MUSCLE_LOG_CRITICALERROR, "Couldn't spawn!\n");
                  return(10);
               }
            }
            else if (strcmp(keyword, "localhost") == 0)
            {
               uint32 ip = Inet_AtoN(value);
               if (ip > 0)
               {
                  LogTime(MUSCLE_LOG_INFO, "IP address [%s] will be used as the localhost address.\n", Inet_NtoA(ip)());
                  SetLocalHostIPOverride(ip);
               }
               else LogTime(MUSCLE_LOG_ERROR, "Error parsing localhost IP address [%s]!\n", value);
            }
            else if (strcmp(keyword, "maxmem") == 0) 
            {
               int megs = muscleMax(1, atoi(value));
               LogTime(MUSCLE_LOG_INFO, "Limiting memory usage to %i megabyte%s.\n", megs, (megs==1)?"":"s");
               maxBytes = megs*1024L*1024L;
            }
            else if (strcmp(keyword, "maxmessagesize") == 0) 
            {
               int k = muscleMax(1, atoi(value));
               LogTime(MUSCLE_LOG_INFO, "Limiting message sizes to %i kilobyte%s.\n", k, (k==1)?"":"s");
               maxMessageSize = k*1024L;
            }
            else if (strcmp(keyword, "maxsendrate") == 0)
            {
               float k = atof(value);
               maxSendRate = muscleMax((uint32)0, (uint32)(k*1024.0f));
            }
            else if (strcmp(keyword, "maxreceiverate") == 0)
            {
               float k = atof(value);
               maxReceiveRate = muscleMax((uint32)0, (uint32)(k*1024.0f));
            }
            else if (strcmp(keyword, "maxcombinedrate") == 0)
            {
               float k = atof(value);
               maxCombinedRate = muscleMax((uint32)0, (uint32)(k*1024.0f));
            }
            else if (strcmp(keyword, "maxnodespersession") == 0) 
            {
               maxNodesPerSession = atoi(value);
               LogTime(MUSCLE_LOG_INFO, "Limiting nodes-per-session to %lu.\n", maxNodesPerSession);
            }
            else if (strcmp(keyword, "maxsessions") == 0) 
            {
               maxSessions = atoi(value);
               LogTime(MUSCLE_LOG_INFO, "Limiting total session count to %lu.\n", maxSessions);
            }
            else if (strcmp(keyword, "maxsessionsperhost") == 0) 
            {
               maxSessionsPerHost = atoi(value);
               LogTime(MUSCLE_LOG_INFO, "Limiting session count for any given host to %lu.\n", maxSessionsPerHost);
            }
            else if (strcmp(keyword, "display") == 0)
            {
               int ll = ParseLogLevelKeyword(value);
               if (ll >= 0) SetConsoleLogLevel(ll);
                       else LogTime(MUSCLE_LOG_INFO, "Error, unknown display log level type [%s]\n", value);
            }
            else if (strcmp(keyword, "log") == 0)
            {
               int ll = ParseLogLevelKeyword(value);
               if (ll >= 0) SetFileLogLevel(ll);
                       else LogTime(MUSCLE_LOG_INFO, "Error, unknown file log level type [%s]\n", value);
            }
            else if (strcmp(keyword, "ban") == 0)
            {
               LogTime(MUSCLE_LOG_INFO, "Banning clients whose IP addresses match [%s]\n", value);   
               bans.AddTail(value);
            }
            else if (strcmp(keyword, "require") == 0)
            {
               LogTime(MUSCLE_LOG_INFO, "Requiring clients whose IP addresses match [%s]\n", value);   
               requires.AddTail(value);
            }
            else if (strncmp(keyword, "priv", 4) == 0)
            {
               const char * privNames[] = {"kick", "ban", "unban", "all"};
               const char * privType = &keyword[4];
               for (int p=0; p<=PR_NUM_PRIVILEGES; p++)  // if (p == PR_NUM_PRIVILEGES), that means all privileges
               {
                  if (strcmp(privType, privNames[p]) == 0)
                  {
                     LogTime(MUSCLE_LOG_INFO, "Clients whose IP addresses match [%s] get %s privileges\n", value, privNames[p]);
                     char tt[32]; 
                     sprintf(tt, "priv%i", p);
                     tempPrivs.AddString(tt, value);
                  }
               }
            }
            else 
            {
               // legacy parsing support
               if ((keyword[0] >= '0')&&(keyword[0] <= '9')) ports.AddTail((uint16)atoi(keyword));
               else LogTime(MUSCLE_LOG_WARNING, "Argument [%s] not parsed.  Run ./muscled help for more info.\n", keyword);
            }
         }
      }
   }

   if ((maxBytes != MUSCLE_NO_LIMIT)&&(usageLimitAllocator)) usageLimitAllocator->SetMaxNumBytes(maxBytes);

   int retVal = 0;
   ReflectServer server(usageLimitAllocator);

   if (maxNodesPerSession != MUSCLE_NO_LIMIT) server.GetCentralState().AddInt32(PR_NAME_MAX_NODES_PER_SESSION, maxNodesPerSession);

   MessageFieldNameIterator iter = tempPrivs.GetFieldNameIterator();
   const String * fn;
   while((fn = iter.GetNextFieldNameString()) != NULL) tempPrivs.CopyName(*fn, server.GetCentralState());

   bool okay = true;

   // If the user asked for bandwidth limiting, create Policy objects to handle that.
   PolicyRef inputPolicyRef, outputPolicyRef;
   if (maxCombinedRate != MUSCLE_NO_LIMIT)
   {
      inputPolicyRef.SetRef(newnothrow RateLimitSessionIOPolicy(maxCombinedRate), NULL);
      outputPolicyRef = inputPolicyRef;
      if (inputPolicyRef()) LogTime(MUSCLE_LOG_INFO, "Limiting aggregate I/O bandwidth to %.02f kilobytes/second.\n", ((float)maxCombinedRate/1024.0f));
      else
      {
         WARN_OUT_OF_MEMORY;
         okay = false;
      }
   }
   else
   {
      if (maxReceiveRate != MUSCLE_NO_LIMIT)
      {
         inputPolicyRef.SetRef(newnothrow RateLimitSessionIOPolicy(maxReceiveRate), NULL);
         if (inputPolicyRef()) LogTime(MUSCLE_LOG_INFO, "Limiting aggregate receive bandwidth to %.02f kilobytes/second.\n", ((float)maxReceiveRate/1024.0f));
         else
         {
            WARN_OUT_OF_MEMORY;
            okay = false;
         }
      }
      if (maxSendRate != MUSCLE_NO_LIMIT)
      {
         outputPolicyRef.SetRef(newnothrow RateLimitSessionIOPolicy(maxSendRate), NULL);
         if (outputPolicyRef()) LogTime(MUSCLE_LOG_INFO, "Limiting aggregate send bandwidth to %.02f kilobytes/second.\n", ((float)maxSendRate/1024.0f)); 
         else
         {
            WARN_OUT_OF_MEMORY;
            okay = false; 
         }
      }
   }
      
   // Set up the Session Factory.  This factory object creates the new StorageReflectSessions
   // as needed when people connect, and also has a filter to keep out the riff-raff.
   StorageReflectSessionFactory factory; factory.SetMaxIncomingMessageSize(maxMessageSize);
   FilterSessionFactory filter(ReflectSessionFactoryRef(&factory, NULL, false), maxSessionsPerHost, maxSessions);
   filter.SetInputPolicy(inputPolicyRef);
   filter.SetOutputPolicy(outputPolicyRef);

   for (int b=bans.GetNumItems()-1;     ((okay)&&(b>=0)); b--) if (filter.PutBanPattern(bans[b]())         != B_NO_ERROR) okay = false;
   for (int a=requires.GetNumItems()-1; ((okay)&&(a>=0)); a--) if (filter.PutRequirePattern(requires[a]()) != B_NO_ERROR) okay = false;

   // Set up ports.  We allow multiple ports, mostly just to show how it can be done;
   // they all get the same set of ban/require patterns (since they all do the same thing anyway).
   if (ports.GetNumItems() == 0) ports.AddTail(DEFAULT_MUSCLED_PORT);
   int numPorts = ports.GetNumItems();
   for (int p=0; p<numPorts; p++)
   {
      if (server.PutAcceptFactory(ports[p], ReflectSessionFactoryRef(&filter, NULL, false)) != B_NO_ERROR)
      {
         LogTime(MUSCLE_LOG_CRITICALERROR, "Error adding port %u, aborting\n", ports[p]);
         okay = false;
         break;
      }
   }

   if (okay)
   {
      retVal = (server.ServerProcessLoop() == B_NO_ERROR) ? 0 : 10;
      if (retVal > 0) LogTime(MUSCLE_LOG_CRITICALERROR, "Server process aborted!\n");
                 else LogTime(MUSCLE_LOG_INFO, "Server process exiting.\n");
   }
   else LogTime(MUSCLE_LOG_CRITICALERROR, "Error occurred during setup, exiting\n");

   server.Cleanup();
   return retVal;
}

#ifdef UNIFIED_DAEMON
int muscledmain(int argc, char ** argv) 
#else
int main(int argc, char ** argv) 
#endif
{
   // Set up memory usage policies for our server.  This is a bit tricky since we
   // are overridding the C++ new and delete operators, but using them to allocate the
   // policy objects at the same time.  So we need to do this as the very first thing.
   UsageLimitProxyMemoryAllocator * usageLimitAllocator   = new UsageLimitProxyMemoryAllocator(MemoryAllocatorRef(new BasicMemoryAllocator, NULL));
   AutoCleanupProxyMemoryAllocator * autoCleanupAllocator = new AutoCleanupProxyMemoryAllocator(MemoryAllocatorRef(usageLimitAllocator, NULL));
   autoCleanupAllocator->AddCallback(OutOfMemoryCallbackRef(new FunctionOutOfMemoryCallback(StorageReflectSession::DrainPools), NULL));
   SetCPlusPlusGlobalMemoryAllocator(MemoryAllocatorRef(autoCleanupAllocator, NULL));

   CompleteSetupSystem css;

   return muscledmainAux(argc, argv, usageLimitAllocator);
};
